<?php
/***********************************************************************
                           
                           Gubed - PHP Debugger                        
                     (c) 2002 - 2004 Linus McCabe                     
                                                                      
 This program is free software; you can redistribute it and/or modify 
 it under the terms of the GNU General Public License as published by 
 the Free Software Foundation; either version 2 of the License, or    
 (at your option) any later version.                                  
                                                                      
 This program is distributed in the hope that it will be useful,      
 but WITHOUT ANY WARRANTY; without even the implied warranty of       
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 GNU General Public License for more details.                         
                                                                      
 You should have received a copy of the GNU General Public License    
 along with this program; if not, write to the Free Software          
 Foundation, Inc.
 59 Temple Place, Suite 330
 Boston, MA  02111-1307  USA
  
************************************************************************/

   /* Initialize gbd globals */
   require_once(dirname(__FILE__).'/GubedGlobals.php');
   if(extension_loaded('tokenizer'))
      require_once(dirname(__FILE__).'/GubedParserTokenizer.php');
   else
      require_once(dirname(__FILE__).'/GubedParserPre42.php');
   
   // This function is called initially and establishes the connection to the debugging server.
   // After handshaking, the name and source (if needed) is sent to the server.
   function gbdInitialize($gbdThisScript, $gbdTimer) {
      restore_error_handler();
  
      $intCnt = 0;
      if(!$GLOBALS['gbd']['gbdSocket']) {
         $GLOBALS['gbd']['gbdSocket'] = @fsockopen($GLOBALS['gbd']['gbdServer'], $GLOBALS['gbd']['gbdPort'], $errno, $errstr, 10);

         if(!$GLOBALS['gbd']['gbdSocket']) {
            echo "Unable to connect to debug server on {$GLOBALS['gbd']['gbdServer']}:{$GLOBALS['gbd']['gbdPort']}. Make sure your client (and proxy if you need it) is started, properly configured and waiting for connections";
            die();
         }
         
         socket_set_timeout($GLOBALS['gbd']['gbdSocket'], 3600, 0);
         socket_set_blocking($GLOBALS['gbd']['gbdSocket'], true);
         $gbdCommand = gbdGetNextCommand($gbdData);
         //$gbdCommand = fgets($GLOBALS['gbd']['gbdSocket'], 256);
         //fgets($GLOBALS['gbd']['gbdSocket'], 256);
         if($gbdCommand == "die")
           die();
         
         socket_set_timeout($GLOBALS['gbd']['gbdSocket'], 5, 0);
          
         gbdSendCommand('sendbreakpoints', '');

         if(!isset($GLOBALS['gbd']['gbdRun'])) {
            $GLOBALS['gbd']['gbdRun'] = false;
            gbdSendCommand('getrunmode', '');
         }
      }

      if($GLOBALS['gbd']['gbdSocket']) {

         if(!isset($GLOBALS['gbd']['gbdPhpOrigCode'][$gbdThisScript]))
            $GLOBALS['gbd']['gbdPhpOrigCode'][$gbdScriptname] = "";
         
         // If this line is activated, run mode will show every single file it enteres (bad performance)
         // But if it's not, wx-client wont update its sourcecodes when changes have been made.. hmmmm
         if(!$GLOBALS['gbd']['gbdRun'])
            gbdSendCommand('initialize', Array('filename' => $gbdThisScript, 'checksum' => md5($GLOBALS['gbd']['gbdPhpOrigCode'][$gbdThisScript])));
//          gbdSendCommand('unparsed', Array('filename' => gbdFixPath($gbdThisScript)));
//          gbdSendCommand('parsetime', Array('filename' => $gbdThisScript, 'time' => sprintf('%.2f', $gbdTimer)));
         //$GLOBALS['gbd']['unparsed'][$gbdThisScript] = true;

      } else {
         print("Unable to connect to debugging socket ($errno): $errstr\n<br>");
         die;
      }

      // set error reporting
      error_reporting(E_ALL);
      set_error_handler("gbdHandleError");

   }


   // This is the core of the debugging. It should be called every other statement.
   // It sends info to the server about what line it's at and asks for further instructions.
   // Then it performs some command or continues the script...
   function gbdGubed($gbdScriptname, $gbdLine, &$gbdVars, $gbdNextInst) {

      if($GLOBALS['gbd']['gbdNoDbg'])
        return 'return true;';

      $gbdTempTimer = gbdGetmicrotime();

      if($gbdLine == "")
         $gbdLine = 0;
      $gbdLine--;

      $GLOBALS['gbd']['gbdPrevScript'] = $gbdScriptname;

      if(count($GLOBALS['gbd']['breakpoints']) > 0
      && isset($GLOBALS['gbd']['breakpoints'][$gbdScriptname])
      && isset($GLOBALS['gbd']['breakpoints'][$gbdScriptname][$gbdLine])) {
         $GLOBALS['gbd']['gbdRun'] = false;
         gbdSendCommand('forcebreak', Array('filename' => gbdFixPath($gbdScriptname), 'line' =>$gbdLine));
      }

      // Backtrace handling
      $gbdFunction = "";
      $gbdClass = "";

      // If our PHP version supports backtrace, get it and build one for displaying
      if(function_exists('debug_backtrace'))
         $gbdBackTrace = debug_backtrace();
      else
         $gbdBackTrace = Array();

      if(isset($gbdBackTrace[1]) &&  isset($gbdBackTrace[1]['function']) &&  $gbdBackTrace[1]['function'] != 'eval') {
         $gbdFunction = $gbdBackTrace[1]['function'];
         if(isset($gbdBackTrace[1]['class']))
            $gbdClass = $gbdBackTrace[1]['class'];

         //if(count($GLOBALS['gbd']['backtrace']) > 0)
            //$GLOBALS['gbd']['backtrace'][count($GLOBALS['gbd']['backtrace']) - 1]['line'] = $gbdBackTrace[1]['line'];
       }
       if(!isset($GLOBALS['gbd']['executiondepth']) || end($GLOBALS['gbd']['executiondepth']) < count($gbdBackTrace))  {
         array_push($GLOBALS['gbd']['backtrace'], 
                          Array(
                              'file' => $gbdScriptname, 
                              'class' => $gbdClass,
                              'function' => $gbdFunction,
                              ));
         $GLOBALS['gbd']['executiondepth'][] = count($gbdBackTrace);
       }

       while(count($GLOBALS['gbd']['executiondepth']) > 0 && end($GLOBALS['gbd']['executiondepth']) > count($gbdBackTrace)) {
         array_pop($GLOBALS['gbd']['backtrace']);
         array_pop($GLOBALS['gbd']['executiondepth']);
       } 
       if(count($GLOBALS['gbd']['backtrace']) > 0)
         $GLOBALS['gbd']['backtrace'][count($GLOBALS['gbd']['backtrace']) - 1]['line'] = $gbdLine + 1;
   
       if(isset($GLOBALS['gbd']['executionhaltdepth']) && end($GLOBALS['gbd']['executiondepth']) <= $GLOBALS['gbd']['executionhaltdepth']) {
         unset($GLOBALS['gbd']['executionhaltdepth']);
         $GLOBALS['gbd']['gbdRun'] = false;
       }
      
      // If we have conditional breakpoints
      $GLOBALS['gbd']['imported'] = false;
      if(count($GLOBALS['gbd']['conditionalbreakpoints']) > 0) {
         // Get the calling function's scope here
         extract($gbdVars);
         $GLOBALS['gbd']['imported'] = true;

         $GLOBALS['gbd']['gbdNoHalt'] = true;
         $GLOBALS['gbd']['gbdNoDbg'] = true;

         // Loop through them and check if they are fulfilled
         foreach ($GLOBALS['gbd']['conditionalbreakpoints'] as $gbdK => $gdbCondition) {

            // If we're in the right function/class
            if(($gdbCondition['function'] == '' || $gdbCondition['function'] == $gbdFunction)
            && ($gdbCondition['class'] == '' || $gdbCondition['class'] == $gbdClass)
            && ($gdbCondition['filename'] == '' || $gdbCondition['filename'] == $gbdScriptname)) {

               // Make sure the expression parses ok
               ob_start();
               $gbdV = eval(" return 1; return {$gdbCondition['expression']};");
               ob_end_clean();
               if($gbdV) {
                  $gbdV  = eval("return {$gdbCondition['expression']};");

                  // Depending on what type of conditional bp this is:
                  if ($gdbCondition['type'] == 'change' && $gbdV != $gdbCondition['value']) {
                  // If the condition was changed, break execution
                     $GLOBALS['gbd']['gbdRun'] = false;
                     $gdbCondition['value'] = strval($gbdV);
                     gbdSendCommand('conditionalbreak', $gdbCondition);
                     gbdSendCommand('showcondition', $gdbCondition);
                  } else if($gdbCondition['type'] != 'change' && $gbdV) {
                     // If the condition is true, break execution
                     $GLOBALS['gbd']['gbdRun'] = false;
                     $gdbCondition['value'] = strval($gbdV);
                     gbdSendCommand('conditionalbreak', $gdbCondition);
                     gbdSendCommand('showcondition', $gdbCondition);
                  } else if(($GLOBALS['gbd']['gbdRun'] == false)){
                     // Breakpoint not fulfilled, If we're not running tell the client...
                     $gdbCondition['value'] = strval($gbdV);
                     gbdSendCommand('showcondition',$gdbCondition);
                  }
                  $GLOBALS['gbd']['conditionalbreakpoints'][$gbdK]['value'] = $gbdV;
               } else {
                  // If they didnt parse, send error msg
                  if($GLOBALS['gbd']['gbdRun'] == false)
                     $gdbCondition['value'] = '-';
                     gbdSendCommand('showcondition', $gdbCondition);
               }
            }
         }
         $GLOBALS['gbd']['gbdNoHalt'] = false;
         $GLOBALS['gbd']['gbdNoDbg'] = false;
      }

      if($GLOBALS['gbd']['gbdRun'] == true) {
      
         // Check for commands every hundredth tick, for somewhat better performance...
         $GLOBALS['gbd']['tickcount']++;
         if($GLOBALS['gbd']['tickcount'] > 100) {
            $gbdCommand = gbdGetNextCommand($gbdData, false);
            $GLOBALS['gbd']['tickcount'] = 0;
         }
         else
            return 'return true;';
            
         if($gbdCommand == "") {
            $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
            return 'return true;';
         }
      }

      // If we're in a script we really shouldnt debug, return immediately
      if(gbdCheckNoDebug($gbdScriptname)) {
         gbdSendCommand("status", Array('message'  => 'Shouldn\'t be here: %1%, %2%',
                                                                  'args' => 2,
                                                                  'arg1' => $gbdScriptname,
                                                                  'arg2' => $gbdLine));
         $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
         return 'return true;';
      }

      $gbdReturnString = "";
      $gbdPrevline = "";
      $gbdMicroTime = gbdGetmicrotime();
      
      gbdSendCommand("commandme", Array('line' =>$gbdLine, 'filename'  => gbdFixPath($gbdScriptname)));

      while(1) {
         if(!isset($gbdCommand))
            $gbdCommand = gbdGetNextCommand($gbdData);
            
         if($gbdCommand != '') {
            switch($gbdCommand) {
               case 'next':
                /****c* Client.to.Script/next
                  * NAME
                  *    next
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to execute the next instruction and proceed to the next one.
                  *     This command will step into function calls and include files. 
                  *
                  * SEE ALSO
                  *     stepover, skip, stepout
                  ****/
                  $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                  return 'return true;';
                  
               case 'stepover':
                /****c* Client.to.Script/stepover
                  * NAME
                  *    stepover
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to execute the next instruction and proceed to the next one.
                  *     This command will not step into function calls and include files.
                  *
                  * SEE ALSO
                  *     next, skip, stepout
                  ****/
                  $GLOBALS['gbd']['executionhaltdepth'] = end($GLOBALS['gbd']['executiondepth']);
                  $GLOBALS['gbd']['gbdRun'] = true;
                  $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                  return 'return true;';

               case 'stepout':
                /****c* Client.to.Script/stepout
                  * NAME
                  *    stepout
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to execute the rest of the instructions in the current function
                  *     or include file and halt again when it reaches the calling/including place.
                  *
                  * SEE ALSO
                  *     next, skip, stepover
                  ****/
                  if(count($GLOBALS['gbd']['executiondepth']) > 1) {
                     end($GLOBALS['gbd']['executiondepth']);
                     $GLOBALS['gbd']['executionhaltdepth'] = prev($GLOBALS['gbd']['executiondepth']);
                  } else
                     $GLOBALS['gbd']['executionhaltdepth'] = end($GLOBALS['gbd']['executiondepth']);
                  $GLOBALS['gbd']['gbdRun'] = true;
                  $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                  return 'return true;';
                  
               case 'pause':
                /****c* Client.to.Script/pause
                  * NAME
                  *    pause
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to halt and await further commands
                  *
                  * SEE ALSO
                  *     trace, run
                  ****/
               case 'trace':
                /****c* Client.to.Script/trace
                  * NAME
                  *    trace
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to constantly tell the client about the current line of execution and
                  *     update watches, etc. As if the user was constantly pressing 'next'.
                  *
                  * SEE ALSO
                  *     pause, run, setdisplaydelay
                  ****/
                  $GLOBALS['gbd']['gbdRun'] = false;
                  break;

               case 'run':
              /****c* Client.to.Script/run
                  * NAME
                  *    run
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to execute as fast as possible. Watches and current position
                  *     won't be updated until the script is paused or it enters trace mode.
                  *
                  * SEE ALSO
                  *     trace, pause
                  ****/ 
                  $GLOBALS['gbd']['gbdRun'] = true;
                  break;

               case 'die':
              /****c* Client.to.Script/die
                  * NAME
                  *    die
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to die. Execution will stop immediately.
                  *
                  * SEE ALSO
                  *     pause
                  ****/ 
                  $GLOBALS['gbd']['buffer'] = ob_get_contents(); 
                  ob_end_clean();
                  die;

               case 'skip':
              /****c* Client.to.Script/skip
                  * NAME
                  *    skip
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to skip the next instruction as if it wasn't there and move on to the next one.
                  *
                  * SEE ALSO
                  *     next, stepover, stepout
                  ****/ 
                  $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                  return 'return false;';

//                case 'conditionalbreakpoint':
//               /****c* Client.to.Script/conditionalbreakpoint
//                   * NAME
//                   *    coconditionalbreakpoint
//                   *
//                   * ARGS
//                   *    * filename - only break if the condition is true in this file (optional)
//                   *    * class - only break if the condition is true and in a method of an object of this class (optional)
//                   *    * function - only break if the condition is true and in a function name like this (optional)
//                   *    * type - either "change" or "true", ie break on change or break when true
//                   *    * expression - what expression to evaluate, for example ($a == 12)
//                   *    * condition - key/id for this bp
//                   *      
//                   * DESCRIPTION
//                   *     Adds a conditional breakpoint
//                   *
//                   *     Note: unlike most other commands, the arguments are separated with a ^, not a colon
//                   *
//                   * SEE ALSO
//                   *     clearconditionalbreakpoint, breakpoint
//                   ****/ 
//                   $GLOBALS['gbd']['gbdNoHalt'] = true;
//                   $GLOBALS['gbd']['gbdNoDbg'] = true;
//                   
//                   // Make sure it parses ok
//                   ob_start();
//                   $gbdV = eval(" return 1; return {$gbdData['expression']};");
//                   ob_end_clean();
//                   if($gbdV) {
//                      $gbdV = eval("return {$gbdData['expression']};");
//                      
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['condition'] =  $gbdData['condition'];
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['filename'] = $gbdData['filename'];
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['class'] =  $gbdData['class'];
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['function'] =  $gbdData['function'];
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['value'] = $gbdV;
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['type'] =  $gbdData['type'];
//                      $GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]['expression'] =  $gbdData['expression'];
// 
//                      gbdSendCommand('showcondition', Array('condition'  => $gbdData['condition'], 'value' => $gbdV));
//                   }
//                   else
//                      gbdSendCommand("status", Array('message'  => 'Invalid expression'));
//                      
//                   $GLOBALS['gbd']['gbdNoHalt'] = false;
//                   $GLOBALS['gbd']['gbdNoDbg'] = false;                     
//                   break;
//                   
//                case 'clearconditionalbreakpoint':
//               /****c* Client.to.Script/clearconditionalbreakpoint
//                   * NAME
//                   *    clearconditionalbreakpoint
//                   *
//                   * ARGS
//                   *    * filename - same as when the breakpoint was added (OBSOLETE?)
//                   *    * class - same as when the breakpoint was added (OBSOLETE?)
//                   *    * function - same as when the breakpoint was added (OBSOLETE?)
//                   *    * type - same as when the breakpoint was added (OBSOLETE?)
//                   *    * expression - same as when the breakpoint was added (OBSOLETE?)
//                   *    * condition - id/key of this bp
//                   *      
//                   * DESCRIPTION
//                   *     Removes a conditional breakpoint
//                   *
//                   *     Note: unlike most other commands, the arguments are separated with a ^, not a colon
//                   *
//                   * SEE ALSO
//                   *     conditionalbreakpoint, clearpoint
//                   ****/ 
//                   unset($GLOBALS['gbd']['conditionalbreakpoints'][$gbdData['condition']]);
//                   break;

               case 'breakpoint':
                  /****c* Client.to.Script/breakpoint
                  * NAME
                  *    breakpoint
                  *
                  * ARGS
                  *    * type - either "line", "change" or "true", ie break on a line, changed expression or when expression is true
                  *    * filename - (conditional/line) for line breakpoints this is the file where the breakpoint should be located/occur, for conditional breakpoints it's an optional filename and halting will only occur if execution is currently in that file
                  *    * line - (line) what line the breakpoint is at
                  *    * class - (conditional) only break if the condition is true and in a method of an object of this class (optional)
                  *    * function - (conditional) only break if the condition is true and in a function name like this (optional)
                  *    * expression - (conditional) what expression to evaluate, for example ($a == 12)
                  *
                  * DESCRIPTION
                  *    breakpoint tells the sctipt to set a breakpoint in the specified
                  *    file on the specified line, or to add a conditional breakpoint with the specified expression.
                  *
                  *    The script might reply with a removebreakpoint command in
                  *    case the breakpoint is not ok
                  * 
                  * SEE ALSO
                  *   removebreakpoint
                  ****/ 
                  
                  if($gbdData['type'] == 'line') {
                     if(!file_exists($gbdData['filename'])) {
                        gbdSendCommand('removebreakpoint', $gbdData); // Return the same array..
                        gbdSendCommand('status', Array('message' => 'File "%1%" does not exist!', 'args' => 1, 'arg1' => $gbdData['filename']));
                     } else {
                        $GLOBALS['gbd']['breakpoints'][$gbdData['filename']][$gbdData['line']] = true;
                     }
                  } else {
                     $GLOBALS['gbd']['gbdNoHalt'] = true;
                     $GLOBALS['gbd']['gbdNoDbg'] = true;
                     
                     // Make sure it parses ok
                     ob_start();
                     $gbdV = eval(" return 1; return {$gbdData['expression']};");
                     ob_end_clean();
                     if($gbdV) {
                        $gbdV = eval("return {$gbdData['expression']};");
                        
                        $gbdData['value'] = strval($gbdV);
                        $GLOBALS['gbd']['conditionalbreakpoints'][] = $gbdData;
                        gbdSendCommand('showcondition', $gbdData);
                        
                     } else {
                        gbdSendCommand('removebreakpoint', $gbdData); // Return the same array..
                        gbdSendCommand("status", Array('message'  => 'Invalid expression'));
                     }
                     $GLOBALS['gbd']['gbdNoHalt'] = false;
                     $GLOBALS['gbd']['gbdNoDbg'] = false;                    
                  }
                  break;

               case 'removebreakpoint':
              /****c* Client.to.Script/removebreakpoint
                  * NAME
                  *    removebreakpoint
                  *
                  * ARGS
                  *    * type - either "line", "change" or "true", ie break on a line, changed expression or when expression is true
                  *    * filename - (conditional/line) for line breakpoints this is the file where the breakpoint should be located/occur, for conditional breakpoints it's an optional filename and halting will only occur if execution is currently in that file
                  *    * line - (line) what line the breakpoint is at
                  *    * class - (conditional) only break if the condition is true and in a method of an object of this class (optional)
                  *    * function - (conditional) only break if the condition is true and in a function name like this (optional)
                  *    * expression - (conditional) what expression to evaluate, for example ($a == 12)
                  *      
                  * DESCRIPTION
                  *     Removes a breakpoint
                  *
                  * SEE ALSO
                  *     breakpoint
                  ****/ 
                  if($gbdData['type'] == 'line') {
                     unset($GLOBALS['gbd']['breakpoints'][$gbdData['filename']][$gbdData['line']]);

                  } else {
                     foreach ($GLOBALS['gbd']['conditionalbreakpoints'] as $gbdK => $gdbCondition) {
                        // If we're in the right function/class
                        if(($gdbCondition['function'] == $gbdData['function'])
                        && ($gdbCondition['expression'] == $gbdData['expression'])
                        && ($gdbCondition['type'] == $gbdData['type'])
                        && ($gdbCondition['class'] == $gbdData['class'])
                        && ($gdbCondition['filename'] == $gbdData['filename'])) {

                           unset($GLOBALS['gbd']['conditionalbreakpoints'][$gbdK]);
                           break;
                        }
                     }
                  }
                  break;

               case 'sendactiveline':
              /****c* Client.to.Script/sendactiveline
                  * NAME
                  *    sendactiveline
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to send the currently active file and line 
                  *
                  * SEE ALSO
                  *
                  ****/ 
                  gbdSendCommand('setactiveline', Array('line' => $gbdLine, 'filename' => gbdFixPath($gbdScriptname)));
                  break;

               case 'sendprotocolversion':
              /****c* Client.to.Script/sendprotocolversion
                  * NAME
                  *    sendprotocolversion
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to send it's protocol version
                  *
                  * SEE ALSO
                  *     protocolversion
                  ****/ 
                  gbdSendCommand('protocolversion', Array('version' => $GLOBALS['gbdProtocolVersion']));
                  break;

               case 'reinitialize':
              /****c* Client.to.Script/reinitialize
                  * NAME
                  *    reinitialize
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to reinitialize, the client can request this action if it was closed
                  *     down and reopened during script execution (can happen when the proxy is used)
                  *     Or if the script tells the client to show a line that it doesent know of (and the source
                  *     can't be loaded form disk)
                  *
                  * SEE ALSO
                  *     initialize
                  ****/ 
                  gbdSendCommand('initialize', Array('filename' => gbdFixPath($gbdScriptname), 'checksum' =>  md5($GLOBALS['gbd']['gbdPhpOrigCode'][$gbdScriptname])));
                  break;

               case 'havesource':
              /****c* Client.to.Script/havesource
                  * NAME
                  *    havesource
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     During initialization, the client tells the script what file is to be entered and
                  *     the client responds with a 'havesource' if it already knows about the source
                  *     of the file. 
                  *
                  * SEE ALSO
                  *     initialize, sendsource
                  ****/ 
                  gbdSendCommand('setactiveline', Array('line' =>$gbdLine, 'filename' => gbdFixPath($gbdScriptname)));
                  break;

               case 'sendsource':
              /****c* Client.to.Script/sendsource
                  * NAME
                  *    sendsource
                  *
                  * ARGS
                  *    * filename - what source is requested
                  *      
                  * DESCRIPTION
                  *     During initialization, the client tells the script what file is to be entered and
                  *     the client responds with a 'sendsource' if it wants the script to send the sourcecode
                  *     of the file. 
                  *
                  * SEE ALSO
                  *     havesource
                  ****/ 
                  if(isset($GLOBALS['gbd']['gbdPhpOrigCode'][$gbdData['filename']])) {
                     gbdSendCommand('source', Array('filename' =>$gbdData['filename'], 'sourcecode' => $GLOBALS['gbd']['gbdPhpOrigCode'][$gbdData['filename']]));
                     gbdSendCommand('setactiveline', Array('line' =>$gbdLine, 'filename' => gbdFixPath($gbdScriptname)));
                  }
                  break;

               case 'setvariable':
              /****c* Client.to.Script/setvariable
                  * NAME
                  *    setvariable
                  *
                  * ARGS
                  *    * variable - varibale to change value of
                  *    * value - new value
                  *
                  * DESCRIPTION
                  *     Tells the script about an expression that needs to be executed. Mainly to set contents of
                  *     variables.
                  *
                  * SEE ALSO
                  *
                  ****/
                  $gbdLine++; // Since we decreased it earlier, wh have to increase it again...

                  // Testig this can give parser errors...
                  ob_start();
                  $gbdTest = eval("return 1; {$gbdData['variable']} =  {$gbdData['value']};");
                  ob_end_clean();

                  if($gbdTest) {
                     $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                     return "\$GLOBALS['gbd']['gbdNoHalt'] = true;
                        {$gbdData['variable']} = {$gbdData['value']};
                        \$GLOBALS['gbd']['gbdNoHalt'] = false;
                        return eval(gbdGubed('{$gbdScriptname}', {$gbdLine}, get_defined_vars(), '{$gbdNextInst}'));";
                  }
                  
                  $GLOBALS['gbd']['gubedtime'] += gbdGetmicrotime() - $gbdTempTimer;
                  return  $gbdData['variable'] . '="' . addslashes($gbdData['value']) . '"; return eval(gbdGubed("'.$gbdScriptname.'",' . $gbdLine . ', get_defined_vars(), "' . $gbdNextInst . '"));';

               case 'getwatch':
              /****c* Client.to.Script/getwatch
                  * NAME
                  *    getwatch
                  *
                  * ARGS
                  *    * variable - expression to watch, for example "$t" to watch the variable "t$"
                  *      
                  * DESCRIPTION
                  *     Requests the value of a watch
                  *
                  * SEE ALSO
                  *     getvariable
                  ****/ 
               case 'getvariable':
              /****c* Client.to.Script/getvariable
                  * NAME
                  *    getvariable
                  *
                  * ARGS
                  *    * variable - expression to get value of, for example "$t"
                  *      
                  * DESCRIPTION
                  *     Requests the value/result of an expression
                  *
                  * SEE ALSO
                  *     getwatch
                  ****/ 
                  if($GLOBALS['gbd']['imported'] == false) {
                     $GLOBALS['gbd']['imported'] = true;
                     extract($gbdVars);
                  }
                     
                  $GLOBALS['gbd']['gbdNoHalt'] = true;
                  $GLOBALS['gbd']['gbdNoDbg'] = true;
                  // In case the user decided to watch $GLOBALS - we hide the gbd ones (Don't wanna send the sources all the time...)
                  $gbdGubedOrigCode = $GLOBALS['gbd']['gbdPhpOrigCode'];
                  $gbdGubedCode = $GLOBALS['gbd']['gbdPhpCode'];
                  
                  unset($GLOBALS['gbd']['gbdPhpOrigCode']);
                  unset($GLOBALS['gbd']['gbdPhpCode']);

                  ob_start();
                  $gbdEval = "
                        \$gbdTemp = {$gbdData['variable']};
                        if(ob_get_length() > 0) {
                          \$gbdOutput = ob_get_contents();
                          ob_clean();
                          echo serialize(Array('return' => \$gbdTemp, 'output' => \$gbdOutput)); 
                        } else {
                          if(!isset(\$gbdTemp))
                            return -1;
                          echo serialize(\$gbdTemp);
                        }
                        return 1;
                     "; 
                  $gbdOk = eval($gbdEval);
                  $gbdValue = ob_get_contents();
                  ob_end_clean();
                  
                  if(!$gbdOk) {
                     $gbdValue = "!";

                  } elseif($gbdOk == -1)
                     $gbdValue = "-";

                  if($gbdCommand == 'getwatch')
                     gbdSendCommand('watch', Array('variable' => "s:".(strlen($gbdData['variable'])).":\"". $gbdData['variable'] ."\";{$gbdValue}"));
                  else
                     gbdSendCommand('variable', Array('variable' => "s:".(strlen($gbdData['variable'])).":\"". $gbdData['variable'] ."\";{$gbdValue}"));
                  
                  // Put the Gubed globals back
                  $GLOBALS['gbd']['gbdPhpOrigCode'] = $gbdGubedOrigCode;
                  $GLOBALS['gbd']['gbdPhpCode'] = $gbdGubedCode;
                  unset($gbdGubedOrigCode);
                  unset($gbdGubedCode);
                  $GLOBALS['gbd']['gbdNoHalt'] = false;
                  $GLOBALS['gbd']['gbdNoDbg'] = false;
                  break;

               case 'processphp':
               /****c* Client.to.Script/processphp
                  * NAME
                  *    processphp
                  *
                  * ARGS
                  *    command - php commands to process
                  *      
                  * DESCRIPTION
                  *     Tells the script to execute a few php commands and return the result
                  *
                  * SEE ALSO
                  *     getwatch, getvariable
                  ****/ 
                   if($GLOBALS['gbd']['imported'] == false) {
                     $GLOBALS['gbd']['imported'] = true;
                     extract($gbdVars);
                  }
                  
                  ob_start();
                  $gbdOutput = eval("return 1; {$gbdData['command']};");
                  ob_end_clean();
                  
                  if($gbdOutput) {
                     
                     $GLOBALS['gbd']['gbdPhpOrigCode']['~'] = $gbdData['command'];
                     gbdSendCommand('initialize', Array('filename' => '~', 'checksum' =>  md5($gbdData['command'])));
                    //gbdSendCommand('source', Array('filename' =>'', 'sourcecode' => $gbdData['command']));
                     
                     $GLOBALS['gbd']['Command'] = '?>' . gbdParseCode('~', "<?php {$gbdData['command']} ?>", false, $gbdTmp, false);
                     $gbdLine++;
                     return "
                        \$GLOBALS['gbd']['gbdRun'] = true;   
                        
                        ob_start();
                        eval(\$GLOBALS['gbd']['Command']);
                        unset(\$gbdData['gbd']['Command']);
                        
                        if(ob_get_length() > 0)
                           \$gbdOutput = ob_get_contents();
                        else
                           \$gbdOutput = '';
                        ob_end_clean();
                        
                        if(\$gbdOutput != '')
                           gbdSendCommand('processoutput', Array('output' => \$gbdOutput));
                     
                        \$GLOBALS['gbd']['gbdRun'] = false;
                        
                        gbdSendCommand('setactiveline', Array('filename' => gbdFixPath('$gbdScriptname'), 'line' => $gbdLine, 'mode' => 'pause'));
                        return eval(gbdGubed('{$gbdScriptname}', {$gbdLine}, get_defined_vars(), '{$gbdNextInst}'));";
                  }
                  else
                     gbdSendCommand('status', Array('message' => "Invalid expression '%1%'", 'args' => 1, 'arg1' => $gbdData['command']));
                  
                  break;
                  
               case 'sendbacktrace':
               /****c* Client.to.Script/sendbacktrace
                  * NAME
                  *    sendbacktrace
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to send the current backtrace
                  *
                  * SEE ALSO
                  *
                  ****/ 
                  $gbdTmp = serialize(array_reverse($GLOBALS['gbd']['backtrace']));
                  gbdSendCommand('backtrace', Array('backtrace' => substr($gbdTmp, 5, strlen($gbdTmp) - 6)));
                  break;
                                    
               case 'sendallvariables':
              /****c* Client.to.Script/sendallvariables
                  * NAME
                  *    sendallvariables
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells the script to send the current scope
                  *
                  * SEE ALSO
                  *
                  ****/ 
                  $gbdVarsCompact = $gbdVars;// gbdArrayCompact($gbdVars);
                  unset($gbdVarsCompact['gbd']);
                  $gbdVarsCompact = serialize($gbdVarsCompact);
                  gbdSendCommand('variable', Array('variable' => "s:".(strlen('Current Scope')).":\"". 'Current Scope' ."\";{$gbdVarsCompact}"));

                  unset($gbdVarsCompact);
                  break;

               case 'sentwatches':
              /****c* Client.to.Script/sentwatches
                  * NAME
                  *    sentwatches
                  *
                  * ARGS
                  *    key - some kind of key that needs to be returned. hmm..
                  *      
                  * DESCRIPTION
                  *     The script confirms that all watches were received
                  *
                  * SEE ALSO
                  *
                  ****/ 
                  gbdSendCommand('sentwatches', Array('key' => $gbdData['key']));
                  break;

               case 'setdisplaydelay':
              /****c* Client.to.Script/setdisplaydelay
                  * NAME
                  *    setdisplaydelay
                  *
                  * ARGS
                  *    * newdelay - new delay in ms to wait between each instruction in trace mode
                  *      
                  * DESCRIPTION
                  *     Set the display delay when in trace mode. Note that setting this to zero does not
                  *     mean trace == run, trace will still be slower due to watch
                  *     updates and other things.
                  *
                  * SEE ALSO
                  *     run
                  ****/ 
                  if(is_numeric($gbdData['newdelay']))
                     $GLOBALS['gbd']['displaydelay'] = $gbdData['newdelay'] * 10000;
                  else
                     gbdSendCommand('status', Array('message' => 'Illegal delay %1%', 'args' => 1, 'arg1' =>  $gbdData['newdelay']));
                  break;

               case 'seterrormask':
              /****c* Client.to.Script/seterrormask
                  * NAME
                  *    seterrormask
                  *
                  * ARGS
                  *    * errormask - the type of errors that should halt execution
                  *      
                  * DESCRIPTION
                  *     Sets the errormask, a combination of
                  *     * E_WARNING (2)
                  *     * E_NOTICE (8)
                  *     * E_USER_ERROR (256)
                  *     * E_USER_WARNING (512)
                  *     * E_USER_NOTICE (1024)
                  *
                  *     Other PHP error types are not handable
                  *
                  * SEE ALSO
                  *     error, http://se.php.net/manual/en/ref.errorfunc.php#errorfunc.constants
                  ****/ 
                  $GLOBALS['gbd']['gbdErrorMask'] = $gbdData['errormask'];
                  break;
                  
               case '':
                  socket_set_timeout($GLOBALS['gbd']['gbdSocket'], 5, 0);
                  break;

               case 'wait':
              /****c* Client.to.Script/wait
                  * NAME
                  *    wait
                  *
                  * ARGS
                  *    none
                  *      
                  * DESCRIPTION
                  *     Tells te script to chill a bit
                  *
                  *     If in trace mode, the client will tell the script to 'wait' then 'next' upon receiving a 'commandme'
                  *
                  * SEE ALSO
                  *
                  ****/ 
                  $gbdMicroTime = $GLOBALS['gbd']['displaydelay'] - (gbdGetmicrotime() - $gbdMicroTime);
                  if($gbdMicroTime > 0) {
                     if(function_exists('usleep'))
                        usleep($gbdMicroTime);
                     else
                        sleep($gbdMicroTime / 1000000);
                  }
                  break;
                  
               default:
                 gbdSendCommand('status', Array('message' => "Unknown command '%1%'", 'args' => 1, 'arg1' => $gbdCommand));
            }
         }
         unset($gbdCommand);
      }
   }


   // This function is called first of all in eval() to indicate that parsing was ok
   function gbdSourceParsed($gbdScriptname) {
      // echo "<br>*** Source parsed: {$GLOBALS['gbd']['unparsed'][$gbdScriptname]} - {gbdScriptname}<br>\n";
      unset($GLOBALS['gbd']['unparsed'][$gbdScriptname]);
   }

   // If error handling is enabled, this function intercepts the error and sends info to the server,
   // then it awaits commands as usual
   function gbdHandleError($number, $text, $filename, $line, $vars) {
      //echo "<br>*** Error: $number, $text, $filename, $line, $vars - {$GLOBALS['gbd']['gbdNoHalt']}<br>\n";
      
      if($GLOBALS['gbd']['gbdNoHalt'])
         return true;

      if($GLOBALS['gbd']['gbdRun'] && !($GLOBALS['gbd']['gbdErrorMask'] & $number))
         return true;

       // Don't want it to say 'error in evaled code'...
      if(dirname($filename) == dirname(__FILE__))
         $filename = $GLOBALS['gbd']['gbdPrevScript'];


      gbdEmptyQueue();
      gbdSendCommand('error', Array('filename' => gbdFixPath($filename), 'line' => $line, 'errnum' => $number, 'errmsg' => $text));

      // Wait for acknowlege
      $gbdCommand = "";
      while(!in_array($gbdCommand, Array("pause", "run", "trace"))) {
         $gbdCommand = gbdGetNextCommand($gbdData);
      }

      if($gbdCommand != "pause")
         return;

      // Turn off runmode if set..
      $GLOBALS['gbd']['gbdRun'] = false;
      if(isset($GLOBALS['gbd']['gbdPhpOrigCode'][$filename]))
         return gbdGubed($filename, $line, $vars, '');

      return;
   }

   // This function is called at forced breakpoints
   function gbdHandleBreakPoint($gbdScriptname, $gbdLine, $gbdVars) {
      
      // Turn off runmode if set...
      $GLOBALS['gbd']['gbdRun'] = false;

      gbdEmptyQueue();
      gbdSendCommand('forcebreak', Array('filename' => gbdFixPath($gbdScriptname), 'line' => $gbdLine - 1));

      // Wait for acknowlege
      $gbdCommand = gbdGetNextCommand($gbdData);
      /*$gbdCommand = trim(fgets($GLOBALS['gbd']['gbdSocket'], 256));
      if($gbdCommand != "")
         $gbdData = trim(fgets($GLOBALS['gbd']['gbdSocket'], 256));*/

      gbdGubed($gbdScriptname, $gbdLine, $gbdVars, '');

      return;
   }

   // Send command
   function gbdSendCommand($gbdCommand, $gbdData) {
      if(is_array($gbdData))
         $gbdData = serialize($gbdData);
      $gbdData = $gbdCommand . ":" . strlen($gbdData) . ';' . $gbdData;
      if(!fwrite($GLOBALS['gbd']['gbdSocket'], $gbdData))
         echo "*** Error sending {$gbdCommand}, socket closed";
   }

   // Read next command from queue
   function gbdGetNextCommand(&$data, $blocking = true) {

      socket_set_blocking($GLOBALS['gbd']['gbdSocket'], false);
   
      // Read all incoming data to buffer non blocking
      $buf = 'x';
      while($buf != '') {
         $buf = fread($GLOBALS['gbd']['gbdSocket'], 8192);
         $GLOBALS['gbd']['gbdBuffer'] .= $buf;
      }

      socket_set_blocking($GLOBALS['gbd']['gbdSocket'], true);
      
      // If we have a full command, return it
      if(gbdCheckDataQueue()) {
         $data = unserialize($GLOBALS['gbd']['gbdData']);
         $cmd = $GLOBALS['gbd']['gbdCommand'];
         
         $GLOBALS['gbd']['gbdData'] = '';
         $GLOBALS['gbd']['gbdCommand'] = '';
         
//          echo "returning {$cmd}<br>";
         return $cmd;
      }
      
      if(!$blocking) {
         $data = Array();
//          echo "returning blank<br>";
         return '';
      }
      
      // Read incoming data in blocking mode
      while(!gbdCheckDataQueue()) 
         $GLOBALS['gbd']['gbdBuffer'] .= fread($GLOBALS['gbd']['gbdSocket'], 8192);
      
      $data = unserialize($GLOBALS['gbd']['gbdData']);
      $cmd = $GLOBALS['gbd']['gbdCommand'];
      
      $GLOBALS['gbd']['gbdData'] = '';
      $GLOBALS['gbd']['gbdCommand'] = '';
      
//       echo "returning {$cmd}<br>";
      return $cmd;
   }
   
      // Check if we have a fully read command
   function gbdCheckDataQueue() {
   
      // First get command
      if($GLOBALS['gbd']['gbdCommand'] == '' && strlen($GLOBALS['gbd']['gbdBuffer']) > 0) {
         $pos = strpos($GLOBALS['gbd']['gbdBuffer'], ';');
         if($pos > 0) {
            $GLOBALS['gbd']['gbdCommand'] = substr($GLOBALS['gbd']['gbdBuffer'], 0, $pos);
            $GLOBALS['gbd']['gbdDataSize'] = substr($GLOBALS['gbd']['gbdCommand'], 1 + strpos($GLOBALS['gbd']['gbdCommand'], ':'));
            $GLOBALS['gbd']['gbdCommand'] = substr($GLOBALS['gbd']['gbdCommand'], 0,  strpos($GLOBALS['gbd']['gbdCommand'], ':'));
            
            $GLOBALS['gbd']['gbdBuffer'] = substr($GLOBALS['gbd']['gbdBuffer'], 1 + $pos);
         }
      }

      // Do we have necessary amount of data?
      if($GLOBALS['gbd']['gbdCommand'] != '' ) {
         if(strlen($GLOBALS['gbd']['gbdBuffer']) >= $GLOBALS['gbd']['gbdDataSize'])  {
            $GLOBALS['gbd']['gbdData'] = substr($GLOBALS['gbd']['gbdBuffer'], 0, $GLOBALS['gbd']['gbdDataSize']);
            $GLOBALS['gbd']['gbdBuffer'] = substr($GLOBALS['gbd']['gbdBuffer'], $GLOBALS['gbd']['gbdDataSize']);
            
            // We have a command and data, return
            return true;
         }
      }
      
      return false;
   }
         
   // Empty the queue
   function gbdEmptyQueue() {
      $gbdCommand = 'x';
      while($gbdCommand != '')
         $gbdCommand = gbdGetNextCommand($gbdData, false);
   }

   // Send info that we've reached the end
   function gbdQuit($gbdThisScript) {
//       gbdSendCommand('end', Array('filename' => gbdFixPath($gbdThisScript)));
   }

   // Send info that we're all done
   function gbdFinalize() {
//       echo "*** Finalize <br>";

      gbdSendCommand('status', Array('message' => 'All done'));
      fflush($GLOBALS['gbd']['gbdSocket']);
      //@fclose($GLOBALS['gbd']['gbdSocket']);
   }

   // Read file from disc
   function gbdReadScript($filename) {
      
      if(function_exists('file_get_contents'))
         return file_get_contents($filename, true);
         
      ob_start();
      $retval = @readfile(gbdFindFile($filename), true);
      if (false !== $retval)
         $retval = ob_get_contents();
      else
         $retval = "Unknown file: {$filename}";
      ob_end_clean();

      return $retval;

   }

   // From php.net, thanks: martijn at grendelman dot net
   // The function below checks the entire include_path for the file, but is altered to 
   // return found filename:
   function gbdFindFile($file) {
      
      if(substr($file,0,1) == DIRECTORY_SEPARATOR && file_exists($file)) 
        return $file;
   
      if(realpath($file) != '')
        return realpath($file);
        
      $incpath = ini_get("include_path");
      foreach (explode(PATH_SEPARATOR, $incpath) as $path) {
         if (file_exists($path.DIRECTORY_SEPARATOR.$file)) 
           return realpath($path.DIRECTORY_SEPARATOR.$file);
      }
      
      return $file;
  }

   function gbdCheckNoDebug($gbdScript) {
      if(!isset($GLOBALS['gbd']['gbdNoDebug']))
         return false;

      if(in_array($gbdScript, $GLOBALS['gbd']['gbdNoDebug']))
         return true;

      reset($GLOBALS['gbd']['gbdNoDebug']);
      while(list($k, $v) = each($GLOBALS['gbd']['gbdNoDebug']))
         if(substr($v, 0, 1) != '#' && preg_match("/$v/i", $gbdScript))
            return true;

      return false;
   }

   // This function does a rewritten header("Location: ...");
   function gbdRelocate($gbdLocation) {
      if(preg_match("/Location:\\s*.*:\\/\\/.*/", $gbdLocation)) {
         Header($gbdLocation);
         return;
      }
      $gbdLocation = str_replace('?', '&', $gbdLocation );
      $gbdLocation ='Location: ' 
                           . $GLOBALS['gbdStartSession'] 
                           . (substr($gbdLocation, strpos($gbdLocation, ':') + 2, 1) == '&' ? $GLOBALS['gbdScript'] : '')
                           . substr($gbdLocation, strpos($gbdLocation, ':') + 2);
       header($gbdLocation);
   }
   
   // Probably obsolete now...
   function gbdFixPath($filepath){
      return $filepath;//str_replace('\\', '/', str_replace(':', '', $filepath));
   }
   
   // From php.net/microtime
   function gbdGetmicrotime(){
      list($usec, $sec) = explode(" ",microtime());
      return ((float)$usec + (float)$sec);
   }

   // Since PHP5 no longer includes $this in get_Defined_vars, we have a trick it by including a reference to the $this,
   // which in turns makes it available.. a bit hackish, perhaps...
   function gbdGet_defined_vars($vars1, $vars2) {
      if(isset($vars2))
         $vars1['gbdthis'] = $vars2;
      return $vars1;   
   }

   // First shutdown function called
   function gbdShutdown1() {
//       echo "*** Shutdown1 <br>";
      
      if(count($GLOBALS['gbd']['unparsed']) > 0) {
         while(list($gbdK, $gbdV) = each($GLOBALS['gbd']['unparsed'])) {
            gbdSendCommand('parsefailed', Array('filename' => gbdFixPath($gbdK)));
         }
      }
      
      // Make sure it's the last one called...
      register_shutdown_function('gbdShutdown2');
   }
   
   // Last shutdownfunction called 
   function gbdShutdown2($output = "!") { 
//       echo "*** Shutdown2 <br>";
      if($output == "!")
      {
         $output = isset($GLOBALS['gbd']['buffer']) ? $GLOBALS['gbd']['buffer'] : '';
         
   //       echo strlen($output);
         if(ob_get_level() > 0) {
            $output .= ob_get_contents();
            ob_end_clean();
         }
      }
//       Header('Content-Type: text/plain');       
//    echo ', ' . strlen($output);
//    echo $output; die();
      // Replace all relative links and actions in $output
      if(strpos(strtolower($output), '<html')) {    
         preg_match_all("/<!--.*-->|<(\\/?[\\w]+)[^>]*>([^<]*)/", $output, $matches, PREG_PATTERN_ORDER);

         $protocol = 'http' . (isset($_SERVER['HTTPS']) && strtolower($_SERVER['HTTPS']) == 'on' ? 's' : '') . '://';

   //        $output = print_r($matches, true);
         $attributes = Array('href', 'usemap', 'src', 'action', 'background', 'cite', 'classid', 'codebase', 'data', 'longdesc', 'profile');
   //       echo "<pre>";
         $addhidden = false;
         for($cnt = 0; $cnt < count($matches[0]); $cnt++) {
   //          echo htmlentities("{$matches[0][$cnt]}");
            foreach($attributes as $attribute) {
               $attribute = strtolower($attribute);
               preg_match_all("/(.*?){$attribute}\\s*=\\s*(\".*?\"|'.*?'|[^\"\'].+?\\s)(.*)/si", $matches[0][$cnt], $submatches, PREG_PATTERN_ORDER);
               if(count($submatches[0]) > 0) {
               
                  for($cnt2 = 0; $cnt2 < count($submatches[0]); $cnt2++) {
                     if(substr($submatches[2][$cnt2], 0, 1) == '"') {
                        $submatches[2][$cnt2] = substr($submatches[2][$cnt2], 1, strlen($submatches[2][$cnt2]) - 2);
                        $quote = '"';
                     } elseif(substr($submatches[2][$cnt2], 0, 1) == "'") {
                        $submatches[2][$cnt2] = substr($submatches[2][$cnt2], 1, strlen($submatches[2][$cnt2]) - 2);
                        $quote = "'";
                     } else
                        $quote = '';

                     if(!preg_match("/^.+:\\/\\//", $submatches[2][$cnt2]) ) {
                        if($attribute == 'action' && $submatches[2][$cnt2] == '') {
                           $submatches[0][$cnt2] = "{$attribute}={$quote}{$protocol}" . $_SERVER['HTTP_HOST'] . $GLOBALS['gbd']['SessionFile'] . $quote;
                           $addhidden  = true;
                        } elseif(substr($submatches[2][$cnt2], 0, 1) == '?') {
                           $submatches[2][$cnt2]  = '&amp;' . substr($submatches[2][$cnt2], 1);
                           $submatches[0][$cnt2] ="{$attribute}={$quote}{$protocol}" . $_SERVER['HTTP_HOST'] . $GLOBALS['gbd']['StartSession'] . $GLOBALS['gbd']['Script'] . $submatches[2][$cnt2] . $quote ;
                        } elseif(preg_match("/.*\\.(html|php|phtml|htm|php.).*/i", $submatches[2][$cnt2])) {
                           $submatches[0][$cnt2] = "{$attribute}={$quote}{$protocol}" . $_SERVER['HTTP_HOST'] . $GLOBALS['gbd']['StartSession'] . $GLOBALS['gbd']['BaseURL'] .str_replace('?', '&', $submatches[2][$cnt2]) . $quote;

                           if($attribute == 'action')
                              $addhidden  = true;
                        } elseif(preg_match("/javascript:.*/i", $submatches[2][$cnt2])) {
                           $submatches[0][$cnt2] = "{$attribute}= " . $quote . $submatches[2][$cnt2] . $quote;
                        } else
                           $submatches[0][$cnt2] = "{$attribute}={$quote}{$protocol}" . $_SERVER['HTTP_HOST'] . $GLOBALS['gbd']['BaseURL'] . $submatches[2][$cnt2] . $quote;
                        $submatches[0][$cnt2] = $submatches[1][$cnt2] . $submatches[0][$cnt2] . $submatches[3][$cnt2];
                     }

                     $matches[0][$cnt] = join('', $submatches[0]);
                     if($addhidden) {
                        $matches[0][$cnt]  .= "<input type=\"hidden\" name=\"gbdScript\" value=\"{$GLOBALS['gbd']['BaseURL']}{$submatches[2][$cnt2]}\" />";
                        $addhidden = false;
                     }
                  }
   //                 $submatches = print_r($submatches, true); 
   //                 echo htmlentities($submatches);
               }
            }
         
         }
      
//       echo htmlentities($output);
//       echo htmlentities(join("", $matches[0]));
//       print_r(headers_sent());
         echo join("", $matches[0]);
      } else
         echo $output;       
   }      
?>
